import asyncio
import importlib
import os
import sys
from pathlib import Path
from typing import Optional

import grpc
from google.protobuf.json_format import Parse, MessageToJson

sys.path.append(str(os.path.dirname(__file__)))

name = 'locapip'
version = __version__ = '0.1.9'

server = grpc.aio.server()
server_task: Optional[asyncio.Task] = None


def import_package(path: str):
    path = str(path)
    sys.path.append(path)

    for p in os.listdir(path):
        if p.endswith('.py') and (Path(path) / p).is_file():
            print(f'import  {Path(path) / p}')
            if p[:-3] in globals():
                importlib.reload(globals()[p[:-3]])
            else:
                importlib.import_module(p[:-3])


def import_package_test():
    import_package(str(Path(os.path.dirname(__file__)) / '_test'))
    print()


async def _serve(port: int):
    import_package_test()
    server.add_insecure_port(f'[::]:{port}')
    await server.start()
    await server.wait_for_termination()


def serve(port: int, run_until_complete=True):
    if run_until_complete:
        asyncio.get_event_loop().run_until_complete(_serve(port))
    else:
        global server_task
        if not server_task:
            server_task.cancel()
        server_task = asyncio.ensure_future(_serve(port))


pb = {}
stub = {}
py_request = {}
py_response = {}


def _py_request_unary(request_type):
    async def foo(cpp_request_):
        return Parse(cpp_request_(str()), request_type())

    return foo


def _py_request_stream(request_type):
    async def foo(cpp_request_):
        while True:
            request_json_ = cpp_request_(str())
            if len(request_json_) == 0:
                break
            yield Parse(request_json_, request_type())

    return foo


async def _py_response_unary(response_message, cpp_response, *args):
    response_json = MessageToJson(response_message, True, True)
    cpp_response(response_json)


async def _py_response_stream(response_message_iterator, cpp_response, *args):
    async for response_message_ in response_message_iterator:
        cpp_response(MessageToJson(response_message_, True, True))
    cpp_response(str())


async def _run(url: str, package: str, rpc: str, py_request_argv, py_response_argv):
    async with grpc.aio.insecure_channel(url) as channel:
        stub_ = getattr(stub[package](channel), rpc)
        if 'UnaryUnary' in str(stub_.__class__):
            py_request_default = _py_request_unary
            py_response_default = _py_response_unary
        elif 'UnaryStream' in str(stub_.__class__):
            py_request_default = _py_request_unary
            py_response_default = _py_response_stream
        elif 'StreamUnary' in str(stub_.__class__):
            py_request_default = _py_request_stream
            py_response_default = _py_response_unary
        elif 'StreamStream' in str(stub_.__class__):
            py_request_default = _py_request_stream
            py_response_default = _py_response_stream

        if package in py_request and rpc in py_request[package]:
            if importlib.import_module(py_request[package][rpc].__module__) is pb[package]:
                py_request_ = py_request_default(py_request[package][rpc])
            else:
                py_request_ = py_request[package][rpc]
        else:
            raise NotImplementedError(f'{package} {rpc} py_request not implemented')

        if package in py_response and rpc in py_response[package]:
            py_response_ = py_response[package][rpc]
        else:
            py_response_ = py_response_default

        if 'UnaryUnary' in str(stub_.__class__):
            response_message = await stub_(await py_request_(*py_request_argv))

        elif 'UnaryStream' in str(stub_.__class__):
            response_message = stub_(await py_request_(*py_request_argv))
        elif 'StreamUnary' in str(stub_.__class__):
            response_message = await stub_(py_request_(*py_request_argv))
        elif 'StreamStream' in str(stub_.__class__):
            response_message = stub_(py_request_(*py_request_argv))
        await py_response_(response_message, *py_response_argv)


def run(url: str, package: str, rpc: str, request, response):
    """
    async run rpc in package on server at url, see specification in each package.py

    :param url: server address
    :param package: protocol buffer package
    :param rpc: remote procedure call
    :param request: vector<function<string<string>>> in py_request
    :param response: vector<function<string<string>>> in py_response
    """

    asyncio.get_event_loop().run_until_complete(_run(url, package, rpc, request, response))



import asyncio
import sys

from google.protobuf.json_format import Parse, MessageToJson

import locapip
import locapip_test_pb2 as pb
import locapip_test_pb2_grpc as pb_rpc


class Service(pb_rpc.TestServicer):
    async def unary_unary(self, request, context):
        print('---------- Test unary_unary ----------')
        print('server receive')
        print(MessageToJson(request, True, True))
        response = pb.Text(text=request.text.replace('request', 'response'))
        print('server send')
        print(MessageToJson(response, True, True))
        return response

    async def unary_stream(self, request, context):
        print('---------- Test unary_stream ----------')
        print('server receive')
        print(MessageToJson(request, True, True))
        for _ in range(3):
            response = pb.Text(text=request.text.replace('request', 'response') + f' stream {_ + 1}')
            print('server send')
            print(MessageToJson(response, True, True))
            yield response

    async def stream_unary(self, request_iterator, context):
        print('---------- Test stream_unary ----------')
        t = []
        async for request in request_iterator:
            t.append(request.text.replace('request', 'response'))
            print('server receive')
            print(MessageToJson(request, True, True))
        response = pb.Text(text=' '.join(t))
        print('server send')
        print(MessageToJson(response, True, True))
        return response

    async def stream_stream(self, request_iterator, context):
        print('---------- Test stream_unary ----------')
        t = []
        async for request in request_iterator:
            t.append(request.text.replace('request', 'response'))
            print('server receive')
            print(MessageToJson(request, True, True))
            response = pb.Text(text=' '.join(t))
            print('server send')
            print(MessageToJson(response, True, True))
            yield response


async def py_request_unary(cpp_request):
    request_message = Parse(cpp_request(str()), pb.Text())
    request_message.text = request_message.text.replace('cpp', 'py')
    print(f'client send')
    print(MessageToJson(request_message, True, True))
    return request_message


async def py_request_stream(cpp_request):
    while True:
        await asyncio.sleep(1)
        request_json_ = cpp_request(str())
        if len(request_json_) == 0:
            break
        request_message = Parse(request_json_, pb.Text())
        request_message.text = request_message.text.replace('cpp', 'py')
        print(f'client send')
        print(MessageToJson(request_message, True, True))
        yield request_message


async def py_response_unary(response_message, cpp_response):
    response_message.text = response_message.text
    print(f'client receive')
    print(MessageToJson(response_message, True, True))
    response_message.text = response_message.text.replace('py', 'cpp')
    cpp_response(MessageToJson(response_message, True, True))


async def py_response_stream(response_message_iterator, cpp_response):
    async for response_message in response_message_iterator:
        response_message.text = response_message.text
        print(f'client receive')
        print(MessageToJson(response_message, True, True))
        response_message.text = response_message.text.replace('py', 'cpp')
        cpp_response(MessageToJson(response_message, True, True))
    cpp_response(str())


pb_rpc.add_TestServicer_to_server(Service(), locapip.server)
locapip.pb[__name__] = pb
locapip.stub[__name__] = pb_rpc.TestStub

locapip.py_request[__name__] = {
    'unary_unary': pb.Text,
    'unary_stream': pb.Text,
    'stream_unary': pb.Text,
    'stream_stream': pb.Text,
}

locapip.py_request[__name__] = {
    'unary_unary': py_request_unary,
    'unary_stream': py_request_unary,
    'stream_unary': py_request_stream,
    'stream_stream': py_request_stream,
}

locapip.py_response[__name__] = {
    'unary_unary': py_response_unary,
    'unary_stream': py_response_stream,
    'stream_unary': py_response_unary,
    'stream_stream': py_response_stream,
}



import logging
from pathlib import Path

import click

import locapip


@click.command()
@click.option('--port', type=int, default=6547)
@click.option('--proto', type=click.Path(file_okay=False, dir_okay=True, path_type=Path), multiple=True)
def main(port: int, proto):
    print(f'{locapip.name} {locapip.version}')
    print(f'--port  {port}')

    if proto is not None:
        for path in proto:
            locapip.import_package(path)

    logging.basicConfig(level=logging.INFO)
    locapip.serve(port)


if __name__ == "__main__":
    main()



import setuptools

from locapip import name, version

import os
from pathlib import Path

long_description = (Path(os.path.dirname(__file__)) / 'README.md').read_text('utf-8')

setuptools.setup(
    name=name,
    version=version,
    author='sjtu_6547',
    author_email='88172828@qq.com',
    description='C++ embedded Python microservices',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://gitlab.com/locapidio/locapip',
    keywords='protobuf grpcio grpcio-tools',
    packages=setuptools.find_packages(),
    include_package_data=True,
    package_data={
        '': ['*.proto'],
    },
    python_requires='>=3.7',
    install_requires=[
        'click>=8', 'grpcio-tools>=1.41.0', ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Framework :: AsyncIO',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
    ],
)
